package edu.berkeley.sbp.misc;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.misc.*;
import edu.berkeley.sbp.tib.*;
import edu.berkeley.sbp.chr.*;
import edu.berkeley.sbp.util.*;

public class RegressionTests {

    public static boolean yes = false;
    public static boolean graph = false;

    public static void main(String[] s) throws Exception {
        try {
            boolean profile = false;
            if (s[0].equals("-graph")) {
                graph = true;
                String[] s2 = new String[s.length-1];
                System.arraycopy(s, 1, s2, 0, s2.length);
                s = s2;
            }
            if (s[0].equals("-profile")) {
                profile = true;
                String[] s2 = new String[s.length-1];
                System.arraycopy(s, 1, s2, 0, s2.length);
                s = s2;
            }

            //MetaGrammar mg0 = new MetaGrammar();
            //mg0.walk(MetaGrammar.meta);
            //System.out.println(mg0);
            System.err.println("parsing " + s[0]);
            Tree<String> res = new CharParser(MetaGrammar.make()).parse(new FileInputStream(s[0])).expand1();
            //System.out.println(mg);
            Union meta = MetaGrammar.make(res, "s");
            System.err.println("parsing " + s[1]);
            SequenceInputStream sis = new SequenceInputStream(new FileInputStream(s[0]), new FileInputStream(s[1]));
            res = new CharParser(meta).parse(sis).expand1();
            Union testcasegrammar = MetaGrammar.make(res, "ts");
            if (testcasegrammar==null) return;
            CharParser parser = new CharParser(testcasegrammar);

            if (profile) {
                System.out.println("\nready...");
                System.in.read();
            }
            System.gc();
            long now = System.currentTimeMillis();
            System.err.println("parsing " + s[2]);
            Forest<String> r2 = parser.parse(new FileInputStream(s[2]));
            System.out.println();
            System.out.println("elapsed = " + (System.currentTimeMillis()-now) + "ms");
            if (profile) {
                System.out.println("\ndone");
                System.in.read();
                System.exit(0);
            }
            System.err.println("expanding...");

            TestCase[] expanded = (TestCase[])new TestCaseBuilder().walk(r2.expand1());
            System.err.println("executing...");
            for(TestCase tc : expanded) {
                tc.execute();
                /*
                String st = "a";
                for(int i=0; i<12; i++) {
                    //System.out.println("length " + st.length());
                    tc.input = st;
                    long nowy = System.currentTimeMillis();
                    GSS.shifts = 0;
                    GSS.reductions = 0;
                    tc.execute();
                    System.out.println("length " + st.length() + " => " + ((System.currentTimeMillis()-nowy)/1000.0) + " " + GSS.shifts + " " + GSS.reductions);
                    st = st+st;
                }
                */
            }

        } catch (Throwable t) {
            System.err.println("\n\nexception thrown, class == " + t.getClass().getName());
            System.err.println(t);
            System.err.println();
            t.printStackTrace();
            System.err.println();
        }
    }

    public static class TestCase {
        private final boolean tib;
        private final boolean jav;
        public /*final*/ String input;        
        public final String[] output;
        public final Union grammar;
        public TestCase(String input, String[] output, Union grammar, boolean tib, boolean jav) throws IOException {
            this.tib = tib;
            this.jav = jav;
            this.input = input;
            this.output = output;
            this.grammar = grammar;
        }
        public String toString() {
            String ret = "testcase {\n" + "  input \""+input+"\";\n";
            for(String s : output) ret += "  output \""+s+"\";\n";
            ret += grammar +"\n}\n";
            return ret;
        }
        public boolean execute() throws Exception {
            Forest<String> res = null;
            ParseFailed pfe = null;
            CharParser parser = new CharParser(grammar);
            //parser.helpgc = false;
            try {
                res = tib 
                    ? /*new CharParser(grammar).parse(new Tib(input))*/ null
                : parser.parse(new StringReader(input));
            } catch (ParseFailed pf) {
                pfe = pf;
            }
            //ystem.out.println("res=="+res);

            if (graph) {
                FileOutputStream fos = new FileOutputStream("out.dot");
                PrintWriter p = new PrintWriter(new OutputStreamWriter(fos));
                GraphViz gv = new GraphViz();
                res.toGraphViz(gv);
                gv.dump(p);
                p.flush();
                p.close();
                System.out.println(parser);
            }
            Collection<Tree<String>> results = res==null ? new HashSet<Tree<String>>() : res.expand(false);
            System.out.print("\r");
            if (results == null || (results.size() == 0 && (output!=null && output.length > 0))) {
                System.out.print("\033[31m");
                System.out.println("PARSE FAILED");
                System.out.print("\033[0m");
                if (pfe != null) pfe.printStackTrace();
            } else {
                System.out.print("\r                                                                                                              \r");
            }
            HashSet<String> outs = new HashSet<String>();
            if (output != null) for(String s : output) outs.add(s.trim());
            boolean bad = false;
            for (Tree<String> r : results) {
                String s = r.toString().trim();
                if (outs.contains(s)) { outs.remove(s); continue; }
                if (!bad) System.out.println(input);
                System.out.print("\033[33m");
                System.out.println("     GOT: " + s);
                bad = true;
            }
            for(String s : outs) {
                if (!bad) System.out.println(input);
                System.out.print("\033[31m");
                System.out.println("EXPECTED: " + s);
                bad = true;
            }
            if (bad) {
                System.out.println("\033[0m");
                return true;
            }             
            System.out.println("\r\033[32mPASS\033[0m                                                                              ");

            return false;
        }
    }

    public static class TestCaseBuilder extends StringWalker {
        public Object walk(Tree<String> tree) {
            try {
                if ("grammar".equals(tree.head())) return MetaGrammar.make(tree, "s");
                else if ("output".equals(tree.head())) return MetaGrammar.string(tree.children());
                else if ("input".equals(tree.head())) return MetaGrammar.string(tree.children());
                else if ("testcase".equals(tree.head())) {
                    String input = MetaGrammar.string(tree.child(0));
                    String[] output = tree.numChildren()>2 ? ((String[])walk(tree, 1)) : new String[0];
                    Union grammar = MetaGrammar.make(tree.child(tree.numChildren()-1), "s");
                    TestCase tc = new TestCase(input, output, grammar, false, false);
                    return tc;
                } else if ("ts".equals(tree.head())) return walk(tree, 0);
                else if (tree.head() == null) {
                    Object[] ret = new Object[tree.numChildren()];
                    for(int i=0; i<ret.length; i++)
                        ret[i] = walk(tree.child(i));
                    return Reflection.lub(ret);
                }
                return super.walk(tree);
            } catch (Exception e) {
                throw new Error(e);
            }
        }
    }
    /*
    public static class JavaGrammar extends MetaGrammar {
        public Object convertLabel(String label) { return new ClassMark(label); }
        public Object convertFieldLabel(String label) { return new FieldMark(label); }

        private static class FieldMark {
            public final String field;
            public FieldMark(String field) { this.field = field; }
            public String toString() { return "."+field; }
        }
        private static class ClassMark {
            public final String clazz;
            public ClassMark(String clazz) { this.clazz = clazz; }
            public String toString() { return clazz+"$"; }
        }

        public static Object build(Tree<Object> t, Class c) throws Exception {
            System.out.println("** build " + c.getName() + " " + t.toPrettyString());
            Object h = t.head();
            if (h instanceof ClassMark) return build2(t, Class.forName(JavaGrammar.class.getName()+"$"+((ClassMark)h).clazz));
            Object o = null;
            if (t.numChildren()==0) o = t.head();
            else if (t.head()==null) return build2(t, c);
            else if (t.head() instanceof FieldMark) {
                return build2(new Tree(null, new Tree[] { t }), c);
            } else {
                throw new Exception("don't know how to cope: " + c.getName() + " <- " + t.toPrettyString());
            }
            return Reflection.rebuild(o, c);
        }
        private static Object build2(Tree<Object> t, Class c) throws Exception {
            System.out.println("** build2 " + c.getName() + " " + t.toPrettyString());
            if (!Reflection.isConcrete(c)) {
                Field f = c.getField("subclasses");
                if (f==null) throw new Exception("don't know how to cope: " + c.getName() + " <- " + t.toPrettyString());
                Class[] subs = (Class[])f.get(null);
                OUTER: for(int i=0; i<subs.length; i++) {
                    for(Tree<Object> t2 : t) {
                        Object o2 = t2.head();
                        System.out.println("checking  " + o2 + " in " + subs[i].getName());
                        if (o2 instanceof FieldMark) {
                            if (subs[i].getField(((FieldMark)o2).field)==null) continue OUTER;
                        }
                    }
                    c = subs[i];
                    break;
                }
            }
            Object o = c.newInstance();
            for(Tree<Object> t2 : t) {
                Object o2 = t2.head();
                if (o2 instanceof FieldMark) {
                    FieldMark f = (FieldMark)o2;
                    Field field = c.getField(ReflectiveWalker.mangle(f.field));
                    Object tgt = build(t2.child(0), field.getType());
                    System.out.println("setting " + f.field +
                                       " on a " + o.getClass().getName() +
                                       " to " + (tgt==null?"null":tgt.getClass().getName()));
                    field.set(o, tgt);
                }
            }
            System.out.println("returning a " + o.getClass().getName());
            return o;
        }
        public static Object build(Tree<Object> t) throws Exception {
            Object h = t.head();
            if (h instanceof ClassMark) {
                ClassMark cm = (ClassMark)h;
                Class c = Class.forName(JavaGrammar.class.getName() + "$" + ReflectiveWalker.mangle(cm.clazz));
                return build2(t, c);
            }
            
            return h==null ? null : h.toString();
        }
        
        public static interface Expr {
            public static Class[] subclasses = new Class[] { _plus_.class, num.class };
        }
        public static class _plus_ implements Expr {
            public Expr left;
            public Expr right;
            public String toString() { return left + "+" + right; }
        }
        public static class num implements Expr {
            public String val;
            public String toString() { return "["+val+"]"; }
        }

    }
    */
    private static String pad(int i,String s) { return s.length() >= i ? s : pad(i-1,s)+" "; }
}

